#!/usr/bin/env perl -w
$|++;

# striplist.pl
# Copyright (C) 2007 by Glenn Tesler, gptesler@math.ucsd.edu
# Last updated: November 8, 2007
#
# This does a brute force enumeration of all arrangements of
# g unsigned permutations, each of size n, where the 1st perm
# is the identity and the others vary.  The "-C" option
# will do the same for "circular" permutations (actually dihedral).
#
# Strip lengths (maximal runs of consecutively ascending or descending #s)
# are determined.  Generating functions are determined from these.
#
# Results should agree with the counts / generating functions
# computed via formulas in countstrips.maple.
#
# striplist.pl [-n #] [-C] [-g #] [-l] [--help]
#   -n #: perm size
#   -C: circular (no -C: linear)
#   -g #: # genomes (default 2)
#   -l: list the arrangements


use strict;
use Data::Dumper;
use Getopt::Long;

# Top level routine to dispatch all the computations and display results
sub countstrips {
    my ($params) = (@_);

    my $n = $params->{'n'};
    my $circ = $params->{'circ'};
    my $nperm1 = $params->{'nperm1'};

    my $n_max = $circ ? $n-1 : $n;   # for circular perms, always put $n first
    my $nf = nfactorial($n_max);
    if ($circ && $n==0) { $nf=1; }
    $params->{'nf'} = $nf;

    $params->{'nstrips'} = {};  # ->{$k} = # arrangements w/exactly $k strips
    $params->{'stype'} = {};    # ->{$s} = # arrangements w/ordered type $s


    ######################################################################
    # Print header
    ######################################################################

    my $lincirc_string = $circ ? "Circular" : "Linear";
    my $show_list = $params->{'list'};
    my $list_string = $show_list ? "On" : "Off";

    print <<"ENDheader";
Examine strip lengths in all arrangements
Permutation size (n):          $n
Number of permutations (g):    $params->{'nperm'}
Permutation type:              $lincirc_string
Show arrangements:             $list_string

ENDheader

    ######################################################################
    # Recursively loop over all arrangements with $nperm1 permutations
    ######################################################################

    gloop($params,$nperm1,[]);

    ######################################################################
    # Display simple counts of # arrangements with each # strips
    ######################################################################

    my $cts = $params->{'nstrips'};
    for (my $i=0; $i<=$n; $i++) {
	my $strip_name = "linear strips";
	my $rearr_name = "arrangements";

	my $ct = $cts->{$i};
	if (!$ct) {
	    $ct = 0;
	} elsif ($ct == 1) {
	    $rearr_name = "arrangement";
	}

	if ($i<=1) {
	    if ($i==1) {
		$strip_name = "linear strip";
	    } elsif ($circ) {
		$strip_name = "linear strips (1 circular strip)";
	    }
	}
	print "$i $strip_name occurs for $ct $rearr_name\n";
    }

    ######################################################################
    # Display ordinary generating function for # arrangements with
    # each # strips
    ######################################################################

    my $maplestr = $circ ? "table_ogf_ac_n_z" : "table_ogf_a_n_z";
    my $g = $nperm1+1;
    $maplestr .=  "[$n,$g] := ";
    my $maplepoly = 
          (join " + ",
	   map { my $c = $cts->{$_}; $c ? "$c*z^$_" : () }
	   ( 0 .. $n ));
    if ($maplepoly eq "") { $maplepoly = "0"; }
    $maplestr .= $maplepoly . ":";
    print "\n", $maplestr, "\n";


    ######################################################################
    # Display table of ordered types & # arrangements with each type
    # With -l option, also print out the arrangements
    ######################################################################

    my $cts_type = $params->{'stype'};
    my $list_type = $params->{'stype_list'};
    if ($circ) {
	print "\nOrdered 1-types:\n";
    } else {
	print "\nOrdered types:\n";
    }

    my @keys = keys %$cts_type;
    if ($circ) {
	@keys =
	    sort
	    map {
		circ_canon_key($_) . "/" . $_
		} @keys;
    } else {
	@keys =
	    sort
	    map {
		sortstrvec($_) . "/" . $_
		} @keys;
    }

    foreach my $sort_key (@keys) {
	my $type_key = $sort_key;
	$type_key =~ s:.*/::;
	my $stype_key = $type_key;

	my $shift = "";
	if ($type_key =~ /(.*),(-\d+)/) {
	    $type_key = $1;
	    $shift = ":" . $2;
	    $stype_key = $type_key . "," . $2;
	}

	if ($show_list) {
	    printf
		"%15s%3s  %10d  %s\n",
		$type_key, $shift,
		$cts_type->{$stype_key},
		(join "; ", @{$list_type->{$stype_key}});
	} else {
	    printf "%15s%3s %10d\n", $type_key, $shift, $cts_type->{$stype_key};
	}
    }

    ######################################################################
    # Display weight generating function for ordered types
    # in syntax for Maple software stripcounts.maple
    ######################################################################

    my $maplestr_ord = maple_otype($params, $cts_type);
    print "\n", $maplestr_ord, "\n";

    ######################################################################
    # Display table of unordered types & # arrangements with each type
    ######################################################################

    print "\nUnordered types:\n";
    my $cts_utype = {};
    foreach my $key (keys %$cts_type) {
	$cts_utype->{sortstrvec($key)} += $cts_type->{$key};
    }

    foreach my $key (sort keys %$cts_utype) {
	printf "%15s %10d\n", $key, $cts_utype->{$key};
    }
    print "\n";

    ######################################################################
    # Display weight generating function for unordered types
    # in syntax for Maple software stripcounts.maple
    ######################################################################

    my $maplestr_utype = maple_utype($params, $cts_utype);
    print $maplestr_utype, "\n\n";

#    print Dumper($params);
}


###############################################################################
# gloop($params, $permno, $perms)
#   Recursively loops over all arrangements
#   $perms = [perm(g-1),perm(g-2),...,]
#       is a list of the permutations generated so far
#       These are actually inverse to the permutations being analyzed
#   $permno = g-1,g-2,...,0 means $permno+1 permutations remain to generate
#   If $permno>0, it recurses to fill in the smaller ones as well.
#   If $permno=0, the last one is implicitly the identity, and
#   the whole arrangement is analyzed.
###############################################################################

sub gloop {
    my ($params, $permno, $perms) = @_;

#    print "gloop: permno=$permno, perms=", Dumper($perms);

    if ($permno == 0) {
	# have generated iterated perms, now count stats for them
	my ($nstrips, $stype) = analyze_perms($params, $perms);
	$params->{'nstrips'}->{$nstrips}++;
	my $stype_str = join ",", @$stype;
	$params->{'stype'}->{$stype_str}++;

	if ($params->{'list'}) {
	    push
		@{$params->{'stype_list'}->{$stype_str}},
		stringify_perms($perms);
	}
    } else {
	my $nf = $params->{'nf'};
	my $n = $params->{'n'};
	my $circ = $params->{'circ'};
	for (my $c=0; $c<$nf; $c++) {
	    my $p = code2perm($c,$n); # unsigned inverse permutation

	    # in circular case, choose a canonical representative thus:
	    # for $n>=1: $p[0] = $n
	    # for $n>=3: $p[1] <= $p[$n-1]
	    next if ($circ && ($n>2 && $p->[1] > $p->[-1]));

	    gloop($params, $permno-1, [@$perms, $p]); # iterate
	}
    }
}

###############################################################################

# string representation of list of permutations
# e.g.,  [3,1,2][1,3,2]
sub stringify_perms {
    my ($perms) = @_;

    join "", map { "[" . (join ",", @{invperm($_)}) . "]" } @$perms;
}

###############################################################################


sub numeric { $a <=> $b };
sub rnumeric { $b <=> $a };

sub sortstrvec {
    my ($s) = @_;
    join ",", map { $_>0 ? $_ : () } (sort rnumeric split ",", $s);
}

# code2perm($c,$n)
#  0 <= $c < $n factorial
# return the permutation of [1,...,$n] coded by $c
sub code2perm {
    my ($c,$n) = (@_);

    my $p = [];
    for (my $i=1; $i<=$n; $i++) {
	my $r = $c % $i;
	$c = int($c / $i);

	# insert $i at position $r
	splice(@$p,$r,0,$i);
    }

    return $p;
}

sub testcode2perm {
    my ($params) = @_;

    my $n = $params->{"n"};
    my $nf = nfactorial($n);
    for (my $i=0; $i<$nf; $i++) {
	my $p = code2perm($i,$n);
	my $q = invperm($p);
	print "$i = [ ", (join ",",@$p), " ] = inv [ ", (join ",",@$q), " ];\n";
    }
    return 0;
}

###############################################################################
# Analyze the strips in the inverse permutation list
#    $perms=[isigma(2),isigma(3),...,isigma(g)]
# relative to isigma(1)=identity.
# isigma(j) is the inverse permutation to sigma(j);
# strip lengths in [sigma(1),...,sigma(g)] are analyzed
# 
#
# Options are set in $params
#
# Return: ($nstrips,$stype)
#    $nstrips = number of linear strips in the arrangement
#    $stype = linear case: string version of ordered type, e.g., '4,1,2'
#             circular case: string version of ordered 1-type + shift:
#             '3,5,1,2': b.p. in sigma1 are
#                 / 1 2 3 / 4 5 6 7 8 / 9 / 10 11 /
#
#             '3,5,1,2,-1': -x @ end means segment with 1 doesn't start @ 1,
#              but is shifted by x.  b.p. in sigma1 are
#                 1 2 / 3 4 5 6 7 / 8 / 9 10 / 11
#             '3,5,1,2,-2':
#                 1 / 2 3 4 5 6 / 7 / 8 9 / 10 11
###############################################################################

sub analyze_perms {
    my ($params, $perms) = @_;
    my $n = $params->{'n'};

    if ($n==0) {
	return (0,[]);
    }

    my $circ = $params->{'circ'};
    my $nperm1 = $params->{'nperm1'};

    my @lens = ();
    my $len = 1;
    for (my $i = 2; $i<=$n; $i++) {
	my $adj = analyze_pos($perms, $circ, $i, $n);
	if ($adj) {
	    $len++;
	} else {
	    push @lens, $len;
	    $len = 1;
	}
    }
    push @lens, $len;

    my $nstrips = scalar(@lens);

    # see if circular & wraparound strip
    if ($circ && scalar(@lens)>1) {
	my $adj = analyze_pos($perms, $circ, 1, $n);
	if ($adj) {
	    # merge last strip length into 1st strip length
	    $lens[0] += $lens[-1];
	    # signify last part is wrap around
	    $lens[-1] = -$lens[-1];
	    $nstrips--;
	}
    }

    if ($circ && $nstrips==1) {
	# In circular case, type [n] is
	# 1 wraparound circular strip, 0 linear strips
	$nstrips = 0;
    }

    return ($nstrips, \@lens);
}

###############################################################################
# Determine if $i-1,$i-2 are adjacent in all permutations.
# Note $perms represents the inverse permutations.
# Takes circularity into account.
# Return: 1 if adjacent, 0 if not
###############################################################################

sub analyze_pos {
    my ($perms, $circ, $i, $n) = @_;
    my ($i1,$i2) = ($i-1,$i-2);
    if ($i<2 || $i>$n) {
	return 0 if (!$circ);
	$i1 = $i1 % $n;
	$i2 = $i2 % $n;
    }
    foreach my $q (@$perms) {
	my $ns = $q->[$i1] - $q->[$i2];
	if (!(abs($ns) == 1 ||
	      ($circ && abs($ns) == $n-1))) {
	    return 0;
	}
    }
    return 1;
}

###############################################################################
# invperm($p)
# Compute inverse of a permutation
###############################################################################

sub invperm {
    my ($p) = @_;
    my $n = scalar(@$p);
    my $q = [];
    for (my $i=0; $i<$n; $i++) {
	$q->[$p->[$i]-1] = $i+1;
    }
    return $q;
}

###############################################################################
# binomial($n,$k)
# Compute binomial coefficient
###############################################################################

sub binomial {
    my ($n,$k) = @_;
    if ($k == 0) { return 1; }
    if ($n < 0 || $k < 0 || $k > $n) { return 0; }

    my $denom = 1;
    my $numer = 1;
    for (my $i=1; $i<=$k; $i++) {
	$numer *= ($n+1-$i);
	$denom *= $i;
    }
    return $numer/$denom;
}

###############################################################################
# nfactorial($n) = 1*2*...*$n
###############################################################################

sub nfactorial {
    my ($n) = @_;

    if ($n<2) {
	return ($n<0) ? 0 : 1
    }

    my $nf = 1;
    for (my $i=1; $i<=$n; $i++) { $nf *= $i; }
    return $nf;
}


###############################################################################
# Maple formula for weight generating function for unordered types
###############################################################################

sub maple_utype {
    my ($params, $cts_utype) = @_;

    my $n = $params->{'n'};
    my $g = $params->{'nperm'};
    my $circ = $params->{'circ'};

    my $maplestr = $circ ? "table_wgf_ac" : "table_wgf_a";
    $maplestr .= "[$n,$g] := ";

    my $first = 1;
    foreach my $key (sort keys %$cts_utype) {
	if ($first) {
	    $first = 0;
	} else {
	    $maplestr .= " + ";
	};
	my $term = $cts_utype->{$key};
	my @vals = split /,/, $key;
	if ($circ && scalar(@vals)<=1) {
		$term .= "*uc[$n]";
	} else {
	    foreach my $val (@vals) {
		$term .= "*u[$val]";
	    }
	}
	$maplestr .= $term;
    }
    if ($first) { $maplestr .= " 0"; }
    $maplestr .= ":";
    return $maplestr;
}

###############################################################################
# Maple formula for weight generating function for ordered types
###############################################################################

sub maple_otype {
    my ($params, $cts_type) = @_;

    my $n = $params->{'n'};
    my $g = $params->{'nperm'};
    my $circ = $params->{'circ'};

    my ($maplestr);
    my $cts_otype;

    if ($circ) {
	$cts_otype = {};
	foreach my $key (keys %$cts_type) {
	    $cts_otype->{circ_canon_key($key)} += $cts_type->{$key};
	}
	$maplestr = "table_wgf_Ac";
    } else {
	$cts_otype = $cts_type;
	$maplestr = "table_wgf_A";
    }

    my @keys = keys %$cts_otype;
    @keys = map { sortstrvec($_) . "/" . $_ } @keys;
    @keys = sort @keys;
#    @keys = map { s:.*/:: } @keys;

    $maplestr .= "[$n,$g] := ";

    my $first = 1;
    foreach my $type_key (@keys) {
	my $key = $type_key;
	$key =~ s:.*/::;
	if ($first) {
	    $first = 0;
	} else {
	    $maplestr .= " + ";
	};
	my $ct = $cts_otype->{$key};

	my $term;
	if (!$circ || $key =~ /,/) {
	    $term = ($key eq "") ? "" : "U[$key]";
	} else {
	    $term = ($key eq "") ? "Uc[0]" : "Uc[$key]";
	}

	$term = ($term eq "") ? $ct : $ct . "*" . $term;
	$maplestr .= $term;
    }
    if ($first) { $maplestr .= " 0"; }
    $maplestr .= ":";
    return $maplestr;
}

###############################################################################
# For types in circular case, determine a canonical cyclic shift
# of the type
###############################################################################

sub circ_canon_key {
    my ($s) = @_;
    my @p = split /,/, $s;

    if (!scalar(@p)) { return ""; }

    if ($p[-1] < 0) { pop @p; }

    my $canon=0;
    my $n = scalar(@p);
    for (my $shift=1; $shift<$n; $shift++) {
	for (my $i=0; $i<$n; $i++) {
	    my $j1 = ($canon + $i) % $n;
	    my $j2 = ($shift + $i) % $n;
	    next if ($p[$j1] == $p[$j2]);
	    if ($p[$j1] < $p[$j2]) {
		$canon = $shift;
	    }
	    last;
	}
    }

    my @q;
    if ($canon == 0) {
	@q = @p;
    } else {
	@q = (@p[$canon .. $n-1], @p[0 .. $canon-1]);
    }
    return join ",", @q;
}

###############################################################################
# Read command line arguments and dispatch computation
###############################################################################

sub main {
    my $params =
      {
       "circ"   => 0, # 0: linear arrangements; 1: circular
       "nperm"  => 2, # number of genomes; 1st is identity
       "nperm1" => 1, # nperm-1 (# perms that vary; to ignore sigma1=identity)
       "n"      => 4, # number of genes
       "list"   => 0, # 0: just count arrangements by type; 1: store & group & list them by type
    };
    my $help;

    GetOptions(
	       "-C" => \ $params->{"circ"},
	       "-g=i" => \ $params->{"nperm"},
	       "-n=i" => \ $params->{"n"},
	       "-l" => \ $params->{"list"},
	       "help" => \$help,
	       );

    if ($help) { print_usage(); exit(0); }

    $params->{'nperm1'} = $params->{'nperm'} - 1;
    countstrips($params);
}

sub print_usage {
    my $usage = qq{
striplist.pl
    [--help]

    [-n #] Permutation size

    [-C] Circular (default linear)

    [-g #] Number of genomes (default 2)

    [-l] List arrangements of each type
         [sigma2][sigma3]...[sigmag]; [sigma2][sigma3]...[sigmag]; ...
         sigma1=identity is implicit and is left out of the display
         Since -l stores all the arrangements, it should only be used
         for small -n and -g.
};
      print $usage;
}

main();
1;
